from typing import List, Union

from copy import deepcopy
from functools import lru_cache
from pathlib import Path

import jsoncomment as jsonc
import numpy as np
import pandas as pd

from shekels.core.config import Config
import shekels.core.data_tools as sdt
# ------------------------------------------------------------------------------


class Database:
    '''
    Database is a class for wrapping a mint transaction DataFrame with a simple
    CRUD-like API. API methods include: update, read and search.
    '''
    @staticmethod
    def from_json(filepath):
        # type: (Union[str, Path]) -> Database
        '''
        Constructs a Database instance from a given JSON filepath.

        Args:
            filepath(Path or str): Path to JSON config file.

        Returns:
            Database: Database instance.
        '''
        with open(filepath) as f:
            config = jsonc.JsonComment().load(f)
        return Database(config)

    def __init__(self, config):
        # type: (dict) -> None
        '''
        Constructs a Database instance.

        Args:
            config (dict): Configuration.
        '''
        config = Config(config)
        config.validate()
        self._config = config.to_primitive()
        self._data = None  # type: Union[None, pd.DataFrame]

    @staticmethod
    def _to_records(data):
        # type: (pd.DataFrame) -> List[dict]
        '''
        Converts given DataFrame to a list of JSONifiable dicts.

        Args:
            data (DataFrame): Data.

        Returns:
            list[dict]: Records.
        '''
        data = data.copy()
        data.date = data.date.apply(lambda x: x.isoformat())
        data = data.replace({np.nan: None}).to_dict(orient='records')
        return data

    @property
    def config(self):
        # type: () -> dict
        '''
        Returns a copy of this instance's configuration.

        Returns:
            dict: Copy of config.
        '''
        return deepcopy(self._config)

    @property
    def data(self):
        # type: () -> Union[None, pd.DataFrame]
        '''
        Returns a copy of this instance's data.

        Returns:
            DataFrame: Copy of data.
        '''
        if self._data is None:
            return None
        return self._data.copy()

    def update(self):
        # type: () -> Database
        '''
        Loads CSV found in config's data_path into self._data.

        Returns:
            Database: self.
        '''
        data = pd.read_csv(self._config['data_path'], index_col=None)
        self._data = sdt.conform(
            data,
            actions=self._config['conform'],
            columns=self._config['columns'],
        )
        return self

    @lru_cache(maxsize=1)
    def read(self):
        # type: () -> List[dict]
        '''
        Returns data if update has been called.

        Raises:
            RuntimeError: If update has not first been called.

        Returns:
            list[dict]: Data as records.
        '''
        if self._data is None:
            msg = 'Database not updated. Please call update.'
            raise RuntimeError(msg)
        return self._to_records(self._data)

    @lru_cache()
    def search(self, query):
        # type: (str) -> List[dict]
        '''
        Search data according to given SQL query.

        Args:
            query (str): SQL query. Make sure to use "FROM data" in query.

        Returns:
            DataFrame: Formatted data.
        '''
        output = sdt.query_data(self._data, query)
        # pandasql coerces Timestamps to strings
        output.date = pd.DatetimeIndex(output.date)
        return self._to_records(output)



from typing import Any

from json import JSONDecodeError
import json

from pandasql import PandaSQLException
from schematics.exceptions import DataError
import flasgger as swg
import flask

from shekels.core.database import Database
import shekels.server.server_tools as svt
# ------------------------------------------------------------------------------


'''
Shekels REST API.
'''


def get_api():
    # type: () -> Any
    '''
    Creates a Blueprint for the Shekels REST API.

    Returns:
        flask.Blueprint: API Blueprint.
    '''
    class ApiBlueprint(flask.Blueprint):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.database = None
            self.config = None
    return ApiBlueprint('api', __name__, url_prefix='')


API = get_api()


@API.route('/api')
def api():
    # type: () -> Any
    '''
    Route to Shekels API documentation.

    Returns:
        html: Flassger generated API page.
    '''
    return flask.redirect(flask.url_for('flasgger.apidocs'))


@API.route('/api/initialize', methods=['POST'])
@swg.swag_from(dict(
    parameters=[
        dict(
            name='config',
            type='string',
            description='Database configuration as JSON string.',
            required=True,
            default='',
        )
    ],
    responses={
        200: dict(
            description='Shekels database successfully initialized.',
            content='application/json',
        ),
        400: dict(
            description='Invalid configuration.',
            example=dict(
                error='''
DataError(
    {'data_path': ValidationError([ErrorMessage("/foo.bar is not in a valid CSV file.", None)])}
)'''[1:],
                success=False,
            )
        )
    }
))
def initialize():
    # type: () -> flask.Response
    '''
    Initialize database with given config.

    Raises:
        RuntimeError: If config is invalid.

    Returns:
        Response: Flask Response instance.
    '''
    config = flask.request.get_json()
    msg = 'Please supply a config dictionary.'
    if config is None:
        raise RuntimeError(msg)

    config = json.loads(config)
    if not isinstance(config, dict):
        raise RuntimeError(msg)

    API.database = Database(config)
    API.config = API.database.config

    return flask.Response(
        response=json.dumps(dict(
            message='Database initialized.',
            config=API.config,
        )),
        mimetype='application/json'
    )


@API.route('/api/update', methods=['POST'])
@swg.swag_from(dict(
    parameters=[],
    responses={
        200: dict(
            description='Shekels database successfully updated.',
            content='application/json',
        ),
        500: dict(
            description='Internal server error.',
        )
    }
))
def update():
    # type: () -> flask.Response
    '''
    Update database.

    Raise:
        RuntimeError: If database has not been initialized.

    Returns:
        Response: Flask Response instance.
    '''
    if API.database is None:
        msg = 'Database not initialized. Please call initialize.'
        raise RuntimeError(msg)

    API.database.update()
    return flask.Response(
        response=json.dumps(dict(
            message='Database updated.',
            config=API.config,
        )),
        mimetype='application/json'
    )


@API.route('/api/read', methods=['GET', 'POST'])
@swg.swag_from(dict(
    responses={
        200: dict(
            description='Read all data from database.',
            content='application/json',
        ),
        500: dict(
            description='Internal server error.',
        )
    }
))
def read():
    # type: () -> flask.Response
    '''
    Read database.

    Raises:
        RuntimeError: If database has not been initilaized.
        RuntimeError: If database has not been updated.

    Returns:
        Response: Flask Response instance.
    '''
    if API.database is None:
        msg = 'Database not initialized. Please call initialize.'
        raise RuntimeError(msg)

    response = {}  # type: Any
    try:
        response = API.database.read()
    except Exception as error:
        return svt.error_to_response(error)

    response = {'response': response}
    return flask.Response(
        response=json.dumps(response),
        mimetype='application/json'
    )


@API.route('/api/search', methods=['POST'])
@swg.swag_from(dict(
    parameters=[
        dict(
            name='query',
            type='string',
            description='SQL query for searching database. Make sure to use "FROM data" in query.',
            required=True,
        )
    ],
    responses={
        200: dict(
            description='Returns a list of JSON compatible dictionaries, one per row.',
            content='application/json',
        ),
        500: dict(
            description='Internal server error.',
        )
    }
))
def search():
    # type: () -> flask.Response
    '''
    Search database with a given SQL query.

    Returns:
        Response: Flask Response instance.
    '''
    params = flask.request.get_json()
    params = json.loads(params)
    try:
        query = params['query']
    except KeyError:
        msg = 'Please supply valid search params in the form '
        msg += '{"query": SQL query}.'
        raise RuntimeError(msg)

    if API.database is None:
        msg = 'Database not initialized. Please call initialize.'
        raise RuntimeError(msg)

    if API.database.data is None:
        msg = 'Database not updated. Please call update.'
        raise RuntimeError(msg)

    response = API.database.search(query)  # type: Any
    response = {'response': response}
    return flask.Response(
        response=json.dumps(response),
        mimetype='application/json'
    )


# ERROR-HANDLERS----------------------------------------------------------------
@API.errorhandler(DataError)
def handle_data_error(error):
    # type: (DataError) -> flask.Response
    '''
    Handles errors raise by config validation.

    Args:
        error (DataError): Config validation error.

    Returns:
        Response: DataError response.
    '''
    return svt.error_to_response(error)


@API.errorhandler(RuntimeError)
def handle_runtime_error(error):
    # type: (RuntimeError) -> flask.Response
    '''
    Handles runtime errors.

    Args:
        error (RuntimeError): Runtime error.

    Returns:
        Response: RuntimeError response.
    '''
    return svt.error_to_response(error)


@API.errorhandler(JSONDecodeError)
def handle_json_decode_error(error):
    # type: (JSONDecodeError) -> flask.Response
    '''
    Handles JSON decode errors.

    Args:
        error (JSONDecodeError): JSON decode error.

    Returns:
        Response: JSONDecodeError response.
    '''
    return svt.error_to_response(error)


@API.errorhandler(PandaSQLException)
def handle_sql_error(error):
    # type: (PandaSQLException) -> flask.Response
    '''
    Handles SQL errors.

    Args:
        error (PandaSQLException): SQL error.

    Returns:
        Response: PandaSQLException response.
    '''
    return svt.error_to_response(error)
# ------------------------------------------------------------------------------


API.register_error_handler(500, handle_data_error)
API.register_error_handler(500, handle_runtime_error)
API.register_error_handler(500, handle_json_decode_error)
API.register_error_handler(500, handle_sql_error)



from typing import Any, Callable

from collections import deque
from copy import deepcopy
from lunchbox.enforce import Enforce
import dash
# ------------------------------------------------------------------------------


class EventListener:
    '''
    Listens for Dash app events and calls registered callbacks.
    '''
    def __init__(self, app, store, memory=10):
        # type: (dash.Dash, dict, int) -> None
        '''
        Constructs EventListener.

        Args:
            app (dash.Dash): Dash application instance.
            store (dict): Dash store.
            memory (int, optional): Number of state changes to remember.
                Default: 10.

        Raises:
            EnforceError: If app is not an instance of dash.Dash.
            EnforceError: If app is not an instance of dict.
            EnforceError: If memory is less than 1.
        '''
        Enforce(app, 'instance of', dash.Dash)
        Enforce(store, 'instance of', dict)
        msg = 'Memory must be greater or equal to {b}. {a} < {b}.'
        Enforce(memory, '>=', 1, message=msg)
        # ----------------------------------------------------------------------

        self._app = app  # type: dash.Dash
        self.events = {}  # type: dict
        self.state = deque([deepcopy(store)], memory)  # type: deque

    def listen(self, event, callback):
        # type: (str, Callable[[Any, dict, dash.Dash], dict]) -> EventListener
        '''
        Listen for given event and call given callback.

        Args:
            event (str): Event name.
            callback (function): Function of form (value, store, app) -> store.

        Raises:
            EnforceError: If event is not a string.

        Returns:
            EventListener: self.
        '''
        msg = 'Event name must be a string. {a} is not a string.'
        Enforce(event, 'instance of', str, message=msg)
        self.events[event] = callback
        return self

    def emit(self, event, value):
        # type: (str, object) -> EventListener
        '''
        Call a registered callback guven an event and value.

        Args:
            event (str): Event name.
            value (object): Value to be given to callback.

        Raises:
            EnforceError: If event is not a string.

        Returns:
            EventListener: self.
        '''
        msg = 'Event name must be a string. {a} is not a string.'
        Enforce(event, 'instance of', str, message=msg)
        if event in self.events:
            store = self.events[event](value, self.store, self._app)
            self.state.append(deepcopy(store))
        return self

    @property
    def store(self):
        # type () -> dict
        '''
        dict: Copy of last item in state.
        '''
        return deepcopy(self.state[-1])



from typing import Any, Dict, List, Tuple, Union

from copy import copy
from copy import deepcopy
from pathlib import Path
import os

from dash.dependencies import Input, Output, State
from flask_caching import Cache
import dash
import dash_core_components as dcc
import dash_html_components as html
import dash_table
import flasgger as swg
import flask
import flask_monitoringdashboard as fmdb
import jsoncomment as jsonc

from shekels.server.api import API
import shekels.core.config as cfg
import shekels.server.components as svc
import shekels.server.event_listener as sev
import shekels.server.server_tools as svt
# ------------------------------------------------------------------------------


'''
Shekels app used for displaying and interacting with database.
'''


def get_app():
    # type: () -> dash.Dash
    '''
    Creates a Shekels app.

    Returns:
        Dash: Dash app.
    '''
    flask_app = flask.Flask('Shekels')
    swg.Swagger(flask_app)
    flask_app.register_blueprint(API)

    # flask monitoring
    fmdb.config.link = 'monitor'
    fmdb.config.monitor_level = 3
    fmdb.config.git = 'https://thenewflesh.github.io/shekels/'
    fmdb.bind(flask_app)

    app = svc.get_dash_app(flask_app)
    app.api = API
    app.client = flask_app.test_client()
    app.cache = Cache(flask_app, config={'CACHE_TYPE': 'SimpleCache'})

    # register event listener
    app.event_listener = sev.EventListener(app, {}) \
        .listen('config-query', svt.config_query_event) \
        .listen('config-search-button', svt.config_query_event) \
        .listen('config-table', svt.config_edit_event) \
        .listen('query', svt.data_query_event) \
        .listen('search-button', svt.data_query_event) \
        .listen('init-button', svt.init_event) \
        .listen('update-button', svt.update_event) \
        .listen('upload', svt.upload_event) \
        .listen('save-button', svt.save_event)
    return app


APP = get_app()


@APP.server.route('/static/<stylesheet>')
def serve_stylesheet(stylesheet):
    # type: (str) -> flask.Response
    '''
    Serve stylesheet to app.

    Args:
        stylesheet (str): stylesheet filename.

    Returns:
        flask.Response: Response.
    '''
    temp = APP.api.config or {}
    color_scheme = copy(cfg.COLOR_SCHEME)
    cs = temp.get('color_scheme', {})
    color_scheme.update(cs)

    params = dict(
        COLOR_SCHEME=color_scheme,
        FONT_FAMILY=temp.get('font_family', cfg.Config.font_family.default),
    )
    content = svt.render_template('style.css.j2', params)
    return flask.Response(content, mimetype='text/css')


# EVENTS------------------------------------------------------------------------
@APP.callback(
    Output('store', 'data'),
    [
        Input('config-query', 'value'),
        Input('config-search-button', 'n_clicks'),
        Input('config-table', 'data'),
        Input('query', 'value'),
        Input('init-button', 'n_clicks'),
        Input('update-button', 'n_clicks'),
        Input('search-button', 'n_clicks'),
        Input('upload', 'contents'),
        Input('save-button', 'n_clicks'),
    ],
    [State('config-table', 'data_previous')]
)
def on_event(*inputs):
    # type: (Tuple[Any, ...]) -> Dict[str, Any]
    '''
    Update database instance, and updates store with input data.

    Args:
        inputs (tuple): Input elements.

    Returns:
        dict: Store data.
    '''
    event = dash.callback_context.triggered[0]['prop_id'].split('.')[0]
    value = dash.callback_context.triggered[0]['value']
    store = APP.event_listener.store

    if event == 'config-table':
        value = dict(new=value[0], old=inputs[-1][0])

    if event == 'update-button' and store.get('/api/initialize') is None:
        APP.event_listener.emit('init-button', None)
    store = APP.event_listener.emit(event, value).store
    return store


@APP.callback(
    Output('plots-content', 'children'),
    [Input('store', 'data')]
)
@APP.cache.memoize(100)
def on_plots_update(store):
    # type: (Dict) -> dash_table.DataTable
    '''
    Updates plots with read information from store.

    Args:
        store (dict): Store data.

    Returns:
        list[dcc.Graph]: Plots.
    '''
    comp = svt.solve_component_state(store)
    if comp is not None:
        return comp
    config = store.get('/config', deepcopy(APP.api.config))
    plots = config.get('plots', [])
    return svc.get_plots(store['/api/search']['response'], plots)


@APP.callback(
    Output('data-content', 'children'),
    [Input('store', 'data')]
)
@APP.cache.memoize(100)
def on_datatable_update(store):
    # type: (Dict) -> dash_table.DataTable
    '''
    Updates datatable with read information from store.

    Args:
        store (dict): Store data.

    Returns:
        DataTable: Dash DataTable.
    '''
    comp = svt.solve_component_state(store)
    if comp is not None:
        return comp
    return svc.get_datatable(store['/api/search']['response'])


@APP.callback(
    Output('config-content', 'children'),
    [Input('store', 'data')]
)
@APP.cache.memoize(100)
def on_config_update(store):
    # type: (Dict[str, Any]) -> List[flask.Response]
    '''
    Updates config table with config information from store.

    Args:
        store (dict): Store data.

    Returns:
        flask.Response: Response.
    '''
    config = store.get('/config', deepcopy(APP.api.config))
    store['/config'] = config
    store['/config/search'] = store.get('/config/search', store['/config'])
    comp = svt.solve_component_state(store, config=True)
    if comp is not None:
        return [
            comp,
            html.Div(className='dummy', children=[
                dash_table.DataTable(id='config-table')
            ])
        ]
    return svc.get_key_value_table(
        store['/config/search'],
        id_='config',
        header='config',
        editable=True,
    )


@APP.callback(
    Output('content', 'children'),
    [Input('tabs', 'value')],
    [State('store', 'data')]
)
def on_get_tab(tab, store):
    # type: (str, Dict) -> Union[flask.Response, List, None]
    '''
    Serve content for app tabs.

    Args:
        tab (str): Name of tab to render.
        store (dict): Store.

    Returns:
        flask.Response: Response.
    '''
    store = store or {}

    if tab == 'plots':
        query = store.get('/api/search/query', APP.api.config['default_query'])
        return svc.get_plots_tab(query)

    elif tab == 'data':
        query = store.get('/api/search/query', APP.api.config['default_query'])
        return svc.get_data_tab(query)

    elif tab == 'config':
        config = store.get('/config', deepcopy(APP.api.config))
        return svc.get_config_tab(config)

    elif tab == 'api':  # pragma: no cover
        return dcc.Location(id='api', pathname='/api')

    elif tab == 'docs':  # pragma: no cover
        return dcc.Location(
            id='docs',
            href='https://thenewflesh.github.io/shekels/'
        )

    elif tab == 'monitor':  # pragma: no cover
        return dcc.Location(id='monitor', pathname='/monitor')
# ------------------------------------------------------------------------------


def run(app, config_path, debug=False, test=False):
    '''
    Runs a given Shekels app.

    Args:
        Dash: Shekels app.
        config_path (str or Path): Path to configuration JSON.
        debug (bool, optional): Whether debug mode is turned on. Default: False.
        test (bool, optional): Calls app.run_server if False. Default: False.
    '''
    config_path = Path(config_path).as_posix()
    with open(config_path) as f:
        config = jsonc.JsonComment().load(f)
    app.api.config = config
    app.api.config_path = config_path
    app.event_listener.state.clear()
    app.event_listener.state.append({})
    if not test:
        app.run_server(debug=debug, host='0.0.0.0', port=80)  # pragma: no cover


if __name__ == '__main__':  # pragma: no cover
    config_path = '/mnt/storage/shekels_config.json'
    if 'REPO_ENV' in os.environ.keys():
        config_path = '/mnt/storage/test_config.json'
    debug = 'DEBUG_MODE' in os.environ.keys()
    run(APP, config_path, debug=debug)



from typing import Any, Dict, List, Optional

from copy import copy
import os

from lunchbox.enforce import Enforce, EnforceError
from pandas import DataFrame, DatetimeIndex
from schematics.exceptions import DataError
import dash
import dash_core_components as dcc
import dash_html_components as html
import dash_table
import flask
import lunchbox.tools as lbt
import rolling_pin.blob_etl as rpb

import shekels.core.config as cfg
import shekels.core.data_tools as sdt
# ------------------------------------------------------------------------------


# TODO: refactor components tests to use selnium and be less brittle
# TODO: add JSON editor component for config
# APP---------------------------------------------------------------------------
def get_dash_app(server, storage_type='memory'):
    # type: (flask.Flask, str) -> dash.Dash
    '''
    Generate Dash Flask app instance.

    Args:
        server (Flask): Flask instance.
        storage_type (str): Storage type (used for testing). Default: memory.

    Returns:
        Dash: Dash app instance.
    '''

    store = dcc.Store(id='store', storage_type=storage_type)

    icon = html.Img(id='icon', src='/assets/icon.svg')
    tabs = dcc.Tabs(
        id='tabs',
        className='tabs',
        value='plots',
        children=[
            dcc.Tab(className='tab', label='plots', value='plots'),
            dcc.Tab(className='tab', label='data', value='data'),
            dcc.Tab(className='tab', label='config', value='config'),
            dcc.Tab(className='tab', label='api', value='api'),
            dcc.Tab(className='tab', label='docs', value='docs'),
            dcc.Tab(className='tab', label='monitor', value='monitor'),
        ],
    )
    tabs = html.Div(id='tabs-container', children=[icon, tabs])

    content = dcc.Loading(
        id="content",
        className='content',
        type="dot",
        fullscreen=True,
    )

    # path to resources inside pip package
    assets = lbt.relative_path(__file__, "../resources")

    # path to resources inside repo
    if 'REPO_ENV' in os.environ.keys():
        assets = lbt.relative_path(__file__, "../../../resources")

    app = dash.Dash(
        name='Shekels',
        title='Shekels',
        server=server,
        external_stylesheets=['/static/style.css'],
        assets_folder=assets,
    )
    app.layout = html.Div(id='layout', children=[store, tabs, content])
    app.config['suppress_callback_exceptions'] = True

    return app


# TABS--------------------------------------------------------------------------
def get_data_tab(query=None):
    # type: (Optional[str]) -> List
    '''
    Get tab element for Shekels data.

    Args:
        query (str, optional): Query string. Default: None.

    Return:
        list: List of elements for data tab.
    '''
    # dummies must go first for element props behavior to work
    content = html.Div(id='lower-content', children=[
        html.Div(id='data-content', className='col', children=[])
    ])
    return [*get_dummy_elements(), get_searchbar(query), content]


def get_plots_tab(query=None):
    # type: (Optional[str]) -> List
    '''
    Get tab element for Shekels plots.

    Args:
        query (str, optional): Query string. Default: None.

    Return:
        list: List of elements for plots tab.
    '''
    # dummies must go first for element props behavior to work
    content = html.Div(id='lower-content', children=[
        html.Div(id='plots-content', className='col', children=[
            dcc.Loading(id="progress-bar", type="circle")
        ])
    ])
    return [*get_dummy_elements(), get_searchbar(query), content]


def get_config_tab(config):
    # type: (Dict) -> List
    '''
    Get tab element for Shekels config.

    Args:
        config (dict): Configuration to be displayed.

    Return:
        list: List of elements for config tab.
    '''
    # dummies must go first for element props behavior to work
    content = html.Div(id='lower-content', children=[
        html.Div(id='config-content', className='col', children=[
            get_key_value_table(
                config, id_='config', header='config', editable=True
            )
        ])
    ])
    return [*get_dummy_elements(), get_configbar(config), content]


# MENUBARS----------------------------------------------------------------------
def get_searchbar(query=None):
    # type: (Optional[str]) -> html.Div
    '''
    Get a row of elements used for querying Shekels data.

    Args:
        query (str, optional): Query string. Default: None.

    Returns:
        Div: Div with query field and buttons.
    '''
    if query is None:
        query = 'select * from data'

    spacer = html.Div(className='col spacer')
    query = dcc.Input(
        id='query',
        className='col query',
        value=query,
        placeholder='SQL query that uses "FROM data"',
        type='text',
        autoFocus=True,
        debounce=True
    )

    search = get_button('search')
    init = get_button('init')
    update = get_button('update')

    row = html.Div(
        className='row',
        children=[query, spacer, search, spacer, init, spacer, update],
    )
    searchbar = html.Div(id='searchbar', className='menubar', children=[row])
    return searchbar


def get_dummy_elements():
    # type: () -> List
    '''
    Returns a list of all elements with callbacks so that the client will not
    throw errors in each tab.

    Returns:
        list: List of html elements.
    '''
    return [
        dcc.Input(className='dummy', id='config-query', value=None),
        html.Div(className='dummy', children=[dash_table.DataTable(id='config-table')]),
        dcc.Input(className='dummy', id='query', value=None),
        html.Div(className='dummy', id='config-search-button', n_clicks=None),
        html.Div(className='dummy', id='search-button', n_clicks=None),
        html.Div(className='dummy', id='init-button', n_clicks=None),
        html.Div(className='dummy', id='update-button', n_clicks=None),
        dcc.Upload(className='dummy', id='upload', contents=None),
        html.Div(className='dummy', id='save-button', n_clicks=None),
    ]


def get_configbar(config, query='select * from config'):
    # type: (Dict, Optional[str]) -> html.Div
    '''
    Get a row of elements used for configuring Shekels.

    Args:
        config (dict): Configuration to be displayed.
        query (str, optional): Query string. Default: None.

    Returns:
        Div: Div with buttons and JSON editor.
    '''
    spacer = html.Div(className='col spacer')
    query = dcc.Input(
        id='config-query',
        className='col query',
        value=query,
        placeholder='SQL query that uses "FROM config"',
        type='text',
        autoFocus=True,
        debounce=True
    )

    search = get_button('search')
    search.id = 'config-search-button'
    init = get_button('init')
    upload = dcc.Upload(
        id='upload',
        children=[get_button('upload')]
    )
    save = get_button('save')
    row = html.Div(
        className='row',
        children=[
            query, spacer, search, spacer, init, spacer, upload, spacer, save
        ],
    )
    configbar = html.Div(id='configbar', className='menubar', children=[row])
    return configbar


# ELEMENTS----------------------------------------------------------------------
def get_button(title):
    # type: (str) -> html.Button
    '''
    Get a html button with a given title.

    Args:
        title (str): Title of button.

    Raises:
        TypeError: If title is not a string.

    Returns:
        Button: Button element.
    '''
    if not isinstance(title, str):
        msg = f'{title} is not a string.'
        raise TypeError(msg)
    return html.Button(id=f'{title}-button', children=[title], n_clicks=0)


def get_key_value_table(
    data, id_='key-value', header='', editable=False, key_order=None
):
    # type (dict, Optional(str), str, bool, Optional(List[str])) -> DataTable
    '''
    Gets a Dash DataTable element representing given dictionary.

    Args:
        data (dict): Dictionary.
        id_ (str, optional): CSS id. Default: 'key-value'.
        header (str, optional): Table header title. Default: ''.
        editable (bool, optional): Whether table is editable. Default: False.
        key_order (list[str], optional): Order in which keys will be displayed.
            Default: None.

    Returns:
        DataTable: Tablular representation of given dictionary.
    '''
    data = rpb.BlobETL(data).to_flat_dict()

    # determine keys
    keys = sorted(list(data.keys()))
    if key_order is not None:
        diff = set(key_order).difference(keys)
        if len(diff) > 0:
            diff = list(sorted(diff))
            msg = f'Invalid key order. Keys not found in data: {diff}.'
            raise KeyError(msg)

        keys = set(keys).difference(key_order)
        keys = sorted(list(keys))
        keys = key_order + keys

    # transform data
    data = [dict(key=k, value=data[k]) for k in keys]

    cols = []  # type: Any
    if len(data) > 0:
        cols = data[0].keys()
    cols = [{'name': x, 'id': x} for x in cols]

    table = dash_table.DataTable(
        data=data,
        data_previous=data,
        columns=cols,
        id=f'{id_}-table',
        sort_action='native',
        sort_mode='multi',
        page_action='none',
        cell_selectable=True,
        editable=editable,
    )
    head = html.Div(className='key-value-table-header', children=header)
    return html.Div(
        id=id_, className='key-value-table-container', children=[head, table]
    )


def get_datatable(data, color_scheme=cfg.COLOR_SCHEME, editable=False):
    # type: (List[Dict], Dict[str, str], bool) -> dash_table.DataTable
    '''
    Gets a Dash DataTable element using given data.
    Assumes dict element has all columns of table as keys.

    Args:
        data (list[dict]): List of dicts.
        color_scheme (dict, optional): Color scheme dictionary.
            Default: COLOR_SCHEME.
        editable (bool, optional): Whether table is editable. Default: False.

    Returns:
        DataTable: Table of data.
    '''
    cs = copy(cfg.COLOR_SCHEME)
    cs.update(color_scheme)

    cols = []  # type: Any
    if len(data) > 0:
        cols = data[0].keys()
    cols = [{'name': x, 'id': x} for x in cols]

    return dash_table.DataTable(
        data=data,
        columns=cols,
        id='datatable',
        fixed_rows=dict(headers=True),
        sort_action='native',
        sort_mode='multi',
        cell_selectable=editable,
        editable=editable,
    )


def get_plots(data, plots):
    # type: (List[dict], List[dict]) -> List[dcc.Graph]
    '''
    Gets a Dash plots using given dicts.
    Assumes dict element has all columns of table as keys.

    Args:
        data (list[dict]): List of dicts defining data.
        plots (list[dict]): List of dicts defining plots.

    Raises:
        EnforceError: If data is not a list of dicts.
        EnforceError: If plots is not a list of dicts.

    Returns:
        list[dcc.Graph]: Plots.
    '''
    msg = 'Data must be a list of dictionaries. Given value: {a}.'
    Enforce(data, 'instance of', list, message=msg)
    for item in data:
        Enforce(item, 'instance of', dict, message=msg)

    msg = 'Plots must be a list of dictionaries. Given value: {a}.'
    Enforce(plots, 'instance of', list, message=msg)
    for item in plots:
        Enforce(item, 'instance of', dict, message=msg)
# --------------------------------------------------------------------------

    data_ = DataFrame(data)
    if 'date' in data_.columns:
        data_.date = DatetimeIndex(data_.date)

    elems = []
    for i, x in enumerate(plots):
        plot = cfg.PlotItem(x)
        plot.validate()
        plot = plot.to_primitive()
        min_width = str(plot['min_width']) + '%'

        try:
            fig = sdt.get_figure(
                data_,
                filters=plot['filters'],
                group=plot['group'],
                pivot=plot['pivot'],
                **plot['figure'],
            )
            fig = dcc.Graph(
                id=f'plot-{i:02d}',
                className='plot',
                figure=fig,
                style={'min-width': min_width},
            )
        except (DataError, EnforceError):
            fig = html.Div(
                id=f'plot-{i:02d}',
                className='plot plot-error',
                style={'min-width': min_width},
                children=html.Div(
                    className='plot-error-container',
                    children=html.Div(
                        className='plot-error-message',
                        children='no data found'
                    )
                )
            )
        elems.append(fig)
    return elems



from typing import Any, Dict, List, Optional, Union

from copy import copy
from random import randint
import datetime as dt
import re

from lunchbox.enforce import Enforce
from pandas import DataFrame, DatetimeIndex
from schematics.exceptions import DataError
import cufflinks as cf  # noqa: F401
import lunchbox.tools as lbt
import numpy as np
import pandasql
import pyparsing as pp
import rolling_pin.blob_etl as rpb
import webcolors

from shekels.core.config import ConformAction
import shekels.core.config as cfg
import shekels.enforce.enforce_tools as eft
# ------------------------------------------------------------------------------


COLOR_COERCION_LUT = {
    '#00CC96': '#5F95DE',
    '#0D0887': '#444459',
    '#19D3F3': '#5F95DE',
    '#242424': '#242424',
    '#276419': '#343434',
    '#2A3F5F': '#444459',
    '#343434': '#343434',
    '#444444': '#444444',
    '#46039F': '#444459',
    '#4D9221': '#444444',
    '#636EFA': '#5F95DE',
    '#7201A8': '#5D5D7A',
    '#7FBC41': '#8BD155',
    '#8E0152': '#444459',
    '#9C179E': '#5D5D7A',
    '#A4A4A4': '#A4A4A4',
    '#AB63FA': '#AC92DE',
    '#B6E880': '#A0D17B',
    '#B6ECF3': '#B6ECF3',
    '#B8E186': '#A0D17B',
    '#BD3786': '#F77E70',
    '#C51B7D': '#F77E70',
    '#C8D4E3': '#B6ECF3',
    '#D8576B': '#F77E70',
    '#DE77AE': '#DE958E',
    '#DE958E': '#DE958E',
    '#E5ECF6': '#F4F4F4',
    '#E6F5D0': '#E9EABE',
    '#EBF0F8': '#F4F4F4',
    '#ED7953': '#F77E70',
    '#EF553B': '#F77E70',
    '#F0F921': '#E8EA7E',
    '#F1B6DA': '#C98FDE',
    '#F4F4F4': '#F4F4F4',
    '#F7F7F7': '#F4F4F4',
    '#FB9F3A': '#EB9E58',
    '#FDCA26': '#EB9E58',
    '#FDE0EF': '#F4F4F4',
    '#FECB52': '#EB9E58',
    '#FF6692': '#F77E70',
    '#FF97FF': '#C98FDE',
    '#FFA15A': '#EB9E58',
}


def conform(data, actions=[], columns=[]):
    # type: (DataFrame, List[dict], List[str]) -> DataFrame
    '''
    Conform given mint transaction data.

    Args:
        data (DataFrame): Mint transactions DataFrame.
        actions (list[dict], optional): List of conform actions. Default: [].
        columns (list[str], optional): List of columns. Default: [].

    Raises:
        DataError: If invalid conform action given.
        ValueError: If source column not found in data columns.

    Returns:
        DataFrame: Conformed DataFrame.
    '''
    for action in actions:
        ConformAction(action).validate()

    data.rename(lbt.to_snakecase, axis=1, inplace=True)
    lut = dict(
        account_name='account',
        transaction_type='type'
    )
    data.rename(lambda x: lut.get(x, x), axis=1, inplace=True)
    data.date = DatetimeIndex(data.date)
    data.amount = data.amount.astype(float)
    data.category = data.category \
        .apply(lambda x: re.sub('&', 'and', lbt.to_snakecase(x)))
    data.account = data.account.apply(lbt.to_snakecase)

    for action in actions:
        source = action['source_column']
        if source not in data.columns:
            msg = f'Source column {source} not found in columns. '
            msg += f'Legal columns include: {data.columns.tolist()}.'
            raise ValueError(msg)

        target = action['target_column']
        if target not in data.columns:
            data[target] = None

        for regex, val in action['mapping'].items():
            if action['action'] == 'overwrite':
                mask = data[source] \
                    .apply(lambda x: re.search(regex, x, flags=re.I)).astype(bool)
                data.loc[mask, target] = val
            elif action['action'] == 'substitute':
                data[target] = data[source] \
                    .apply(lambda x: re.sub(regex, val, str(x), flags=re.I))

    if columns != []:
        data = data[columns]
    return data


def filter_data(data, column, comparator, value):
    # type: (DataFrame, str, str, Any) -> DataFrame
    '''
    Filters given data via comparator(column value, value).

    Legal comparators:

        * == ``lambda a, b: a == b``
        * != ``lambda a, b: a != b``
        *  > ``lambda a, b: a > b``
        * >= ``lambda a, b: a >= b``
        *  < ``lambda a, b: a < b``
        * =< ``lambda a, b: a <= b``
        *  ~ ``lambda a, b: bool(re.search(a, b, flags=re.I))``
        * !~ ``lambda a, b: not bool(re.search(a, b, flags=re.I))``

    Args:
        data (DataFrame): DataFrame to be filtered.
        column (str): Column name.
        comparator (str): String representation of comparator.
        value (object): Value to be compared.

    Raises:
        EnforceError: If data is not a DataFrame.
        EnforceError: If column is not a string.
        EnforceError: If column not in data columns.
        EnforceError: If illegal comparator given.
        EnforceError: If comparator is ~ or !~ and value is not a string.

    Returns:
        DataFrame: Filtered data.
    '''
    Enforce(data, 'instance of', DataFrame)
    msg = 'Column must be a str. {a} is not str.'
    Enforce(column, 'instance of', str, message=msg)
    eft.enforce_columns_in_dataframe([column], data)

    lut = {
        '==': lambda a, b: a == b,
        '!=': lambda a, b: a != b,
        '>': lambda a, b: a > b,
        '>=': lambda a, b: a >= b,
        '<': lambda a, b: a < b,
        '<=': lambda a, b: a <= b,
        '~': lambda a, b: bool(re.search(b, a, flags=re.I)),
        '!~': lambda a, b: not bool(re.search(b, a, flags=re.I)),
    }
    msg = 'Illegal comparator. {a} not in [==, !=, >, >=, <, <=, ~, !~].'
    Enforce(comparator, 'in', lut.keys(), message=msg)

    if comparator in ['~', '!~']:
        msg = 'Value must be string if comparator is ~ or !~. {a} is not str.'
        Enforce(value, 'instance of', str, message=msg)
    # --------------------------------------------------------------------------

    op = lut[comparator]
    mask = data[column].apply(lambda x: op(x, value))
    data = data[mask]
    return data


def group_data(data, columns, metric, datetime_column='date'):
    # type: (DataFrame, Union[str, List[str]], str, str) -> DataFrame
    '''
    Groups given data by given columns according to given metric.
    If a legal time interval is given in the columns, then an additional special
    column of that same name is added to the data for grouping.

    Legal metrics:

        *   max ``lambda x: x.max()``
        *  mean ``lambda x: x.mean()``
        *   min ``lambda x: x.min()``
        *   std ``lambda x: x.std()``
        *   sum ``lambda x: x.sum()``
        *   var ``lambda x: x.var()``
        * count ``lambda x: x.count()``

    Legal time intervals:

        * year
        * quarter
        * month
        * two_week
        * week
        * day
        * hour
        * half_hour
        * quarter_hour
        * minute
        * second
        * microsecond

    Args:
        data (DataFrame): DataFrame to be grouped.
        columns (str or list[str]): Columns to group data by.
        metric (str): String representation of metric.
        datetime_column (str, optinal): Datetime column for time grouping.
            Default: date.

    Raises:
        EnforceError: If data is not a DataFrame.
        EnforceError: If columns not in data columns.
        EnforceError: If illegal metric given.
        EnforceError: If time interval in columns and datetime_column not in
            columns.

    Returns:
        DataFrame: Grouped data.
    '''
    # luts
    met_lut = {
        'max': lambda x: x.max(),
        'mean': lambda x: x.mean(),
        'min': lambda x: x.min(),
        'std': lambda x: x.std(),
        'sum': lambda x: x.sum(),
        'var': lambda x: x.var(),
        'count': lambda x: x.count(),
    }

    time_lut = {
        'year': lambda x: dt.datetime(x.year, 1, 1),
        'quarter': lambda x: dt.datetime(
            x.year, int(np.ceil(x.month / 3) * 3 - 2), 1
        ),
        'month': lambda x: dt.datetime(x.year, x.month, 1),
        'two_week': lambda x: dt.datetime(
            x.year, x.month, min(int(np.ceil(x.day / 14) * 14 - 13), 28)
        ),
        'week': lambda x: dt.datetime(
            x.year, x.month, max(1, min([int(x.month / 7) * 7, 28]))
        ),
        'day': lambda x: dt.datetime(x.year, x.month, x.day),
        'hour': lambda x: dt.datetime(x.year, x.month, x.day, x.hour),
        'half_hour': lambda x: dt.datetime(
            x.year, x.month, x.day, x.hour, int(x.minute / 30) * 30
        ),
        'quarter_hour': lambda x: dt.datetime(
            x.year, x.month, x.day, x.hour, int(x.minute / 15) * 15
        ),
        'minute': lambda x: dt.datetime(
            x.year, x.month, x.day, x.hour, x.minute
        ),
        'second': lambda x: dt.datetime(
            x.year, x.month, x.day, x.hour, x.minute, x.second
        ),
        'microsecond': lambda x: dt.datetime(
            x.year, x.month, x.day, x.hour, x.minute, x.second, x.microsecond
        ),
    }
    # --------------------------------------------------------------------------

    # enforcements
    Enforce(data, 'instance of', DataFrame)
    columns_ = columns  # type: Any
    if type(columns_) != list:
        columns_ = [columns_]

    cols = list(filter(lambda x: x not in time_lut.keys(), columns_))
    eft.enforce_columns_in_dataframe(cols, data)

    msg = '{a} is not a legal metric. Legal metrics: {b}.'
    Enforce(metric, 'in', sorted(list(met_lut.keys())), message=msg)

    # time column
    if len(columns_) > len(cols):
        eft.enforce_columns_in_dataframe([datetime_column], data)
        msg = 'Datetime column of type {a}, it must be of type {b}.'
        Enforce(
            data[datetime_column].dtype.type, '==', np.datetime64, message=msg
        )
    # --------------------------------------------------------------------------

    for col in columns_:
        if col in time_lut.keys():
            op = time_lut[col]
            data[col] = data[datetime_column].apply(op)
    agg = met_lut[metric]
    cols = data.columns.tolist()
    grp = data.groupby(columns_, as_index=False)
    output = agg(grp)

    # get first value for columns that cannot be computed by given metric
    diff = set(cols).difference(output.columns.tolist())
    if len(diff) > 0:
        first = grp.first()
        for col in diff:
            output[col] = first[col]
    return output


def pivot_data(data, columns, values=[], index=None):
    # type: (DataFrame, List[str], List[str], Optional[str]) -> DataFrame
    '''
    Pivots a given dataframe via a list of columns.

    Legal time columns:

        * date
        * year
        * quarter
        * month
        * two_week
        * week
        * day
        * hour
        * half_hour
        * quarter_hour
        * minute
        * second
        * microsecond

    Args:
        data (DataFrame): DataFrame to be pivoted.
        columns (list[str]): Columns whose unique values become separate traces
            within a plot.
        values (list[str], optional): Columns whose values become the values
            within each trace of a plot. Default: [].
        index (str, optional): Column whose values become the y axis values of a
            plot. Default: None.

    Raises:
        EnforceError: If data is not a DataFrame.
        EnforceError: If data is of zero length.
        EnforceError: If columns not in data columns.
        EnforceError: If values not in data columns.
        EnforceError: If index not in data columns or legal time columns.

    Returns:
        DataFrame: Pivoted data.
    '''
    time_cols = [
        'date', 'year', 'quarter', 'month', 'two_week', 'week', 'day', 'hour',
        'half_hour', 'quarter_hour', 'minute', 'second', 'microsecond',
    ]

    Enforce(data, 'instance of', DataFrame)
    msg = 'DataFrame must be at least 1 in length. Given length: {a}.'
    Enforce(len(data), '>=', 1, message=msg)
    eft.enforce_columns_in_dataframe(columns, data)
    eft.enforce_columns_in_dataframe(values, data)
    if index is not None:
        msg = '{a} is not in legal column names: {b}.'
        Enforce(index, 'in', data.columns.tolist() + time_cols, message=msg)
    # --------------------------------------------------------------------------

    vals = copy(values)
    if index is not None and index not in values:
        vals.append(index)

    if index in time_cols:
        data[index] = data[index] \
            .apply(lambda x: x + dt.timedelta(microseconds=randint(0, 999999)))

    data = data.pivot(columns=columns, values=vals, index=index)
    data = data[values]
    data.columns = data.columns.droplevel(0)
    return data


def get_figure(
    data,              # type: DataFrame
    filters=[],        # type: List[dict]
    group=None,        # type: Optional[dict]
    pivot=None,        # type: Optional[dict]
    kind='bar',        # type: str
    color_scheme={},   # type: Dict[str, str]
    x_axis=None,       # type: Optional[str]
    y_axis=None,       # type: Optional[str]
    title=None,        # type: Optional[str]
    x_title=None,      # type: Optional[str]
    y_title=None,      # type: Optional[str]
    bins=50,           # type: int
    bar_mode='stack',  # type: str
):
    '''
    Generates a plotly figure dictionary from given data and manipulations.

    Args:
        data (DataFrame): Data.
        filters (list[dict], optional): List of filters for data. Default: [].
        group (dict, optional): Grouping operation. Default: None.
        pivot (dict, optional): Pivot operation. Default: None.
        kind (str, optional): Kind of plot. Default: bar.
        color_scheme (dict[str, str], optional): Color scheme. Default: {}.
        x_axis (str): Column to use as x axis: Default: None.
        y_axis (str): Column to use as y axis: Default: None.
        title (str, optional): Title of plot. Default: None.
        x_title (str, optional): Title of x axis. Default: None.
        y_title (str, optional): Title of y axis. Default: None.
        bins (int, optional): Number of bins if histogram. Default: 50.
        bar_mode (str, optional): How bars in bar graph are presented.
            Default: stack.

    Raises:
        DataError: If any filter in filters is invalid.
        DataError: If group is invalid.
        DataError: If pivot is invalid.

    Returns:
        dict: Plotly Figure as dictionary.
    '''
    data = data.copy()

    # filter
    for f in filters:
        f = cfg.FilterAction(f)
        try:
            f.validate()
        except DataError as e:
            raise DataError({'Invalid filter': e.to_primitive()})

        f = f.to_primitive()
        if len(data) == 0:
            break
        data = filter_data(data, f['column'], f['comparator'], f['value'])

    # group
    if group is not None:
        grp = group  # type: Any
        grp = cfg.GroupAction(grp)
        try:
            grp.validate()
        except DataError as e:
            raise DataError({'Invalid group': e.to_primitive()})
        grp = grp.to_primitive()

        data = group_data(
            data,
            grp['columns'],
            grp['metric'],
            datetime_column=grp['datetime_column'],
        )

    # pivot
    if pivot is not None:
        pvt = pivot  # type: Any
        pvt = cfg.PivotAction(pvt)
        try:
            pvt.validate()
        except DataError as e:
            raise DataError({'Invalid pivot': e.to_primitive()})
        pvt = pvt.to_primitive()

        data = pivot_data(
            data, pvt['columns'], values=pvt['values'], index=pvt['index']
        )

    # create figure
    figure = data.iplot(
        kind=kind, asFigure=True, theme='henanigans', colorscale='henanigans',
        x=x_axis, y=y_axis, title=title, xTitle=x_title, yTitle=y_title,
        barmode=bar_mode, bins=bins
    ).to_dict()  # type: dict
    figure['layout']['title']['font']['color'] = '#F4F4F4'
    figure['layout']['xaxis']['title']['font']['color'] = '#F4F4F4'
    figure['layout']['yaxis']['title']['font']['color'] = '#F4F4F4'
    if color_scheme != {}:
        figure = conform_figure(figure, color_scheme)
    return figure


def parse_rgba(string):
    '''
    Parses rgb and rgba strings into tuples of numbers.

    Example:
        >>>parse_rgba('rgb(255, 0, 0)')
        (255, 0, 0)
        >>>parse_rgba('rgba(255, 0, 0, 0.5)')
        (255, 0, 0, 0.5)
        >>>parse_rgba('foo') is None
        True

    Args:
        string (str): String to be parsed.

    Returns:
        tuple: (red, green, blue) or (red, green, blue, alpha)
    '''
    result = re.search(r'rgba?\((\d+, \d+, \d+(, \d+\.?\d*)?)\)', string)
    if result is None:
        return None

    result = result.group(1)
    result = re.split(', ', result)
    if len(result) == 3:
        result = tuple(map(int, result))
        return result

    result = list(map(int, result[:-1])) + [float(result[-1])]
    result = tuple(result)
    return result


def conform_figure(figure, color_scheme):
    '''
    Conforms given figure to use given color scheme.

    Args:
        figure (dict): Plotly figure.
        color_scheme (dict): Color scheme dictionary.

    Returns:
        dict: Conformed figure.
    '''
    # create hex to hex lut
    lut = {}
    for key, val in cfg.COLOR_SCHEME.items():
        if key in color_scheme:
            lut[val] = color_scheme[key]

    # rgba? to hex --> coerce to standard colors --> coerce with color_scheme
    figure = rpb.BlobETL(figure) \
        .set(
            predicate=lambda k, v: isinstance(v, str) and 'rgb' in v,
            value_setter=lambda k, v: webcolors.rgb_to_hex(parse_rgba(v)[:3]).upper()) \
        .set(
            predicate=lambda k, v: isinstance(v, str),
            value_setter=lambda k, v: COLOR_COERCION_LUT.get(v, v)) \
        .set(
            predicate=lambda k, v: isinstance(v, str),
            value_setter=lambda k, v: lut.get(v, v)) \
        .to_dict()
    return figure


# SQL-PARSING-------------------------------------------------------------------
def get_sql_grammar():
    '''
    Creates a grammar for parsing SQL queries.

    Returns:
       MatchFirst: SQL parser.
    '''
    select = pp.Regex('select', flags=re.I) \
        .setParseAction(lambda s, l, t: 'select') \
        .setResultsName('operator')
    from_ = pp.Suppress(pp.Regex('from', flags=re.I))
    table = (from_ + pp.Regex('[a-z]+', flags=re.I)) \
        .setParseAction(lambda s, l, t: t[0]) \
        .setResultsName('table')
    regex = pp.Regex('~|regex').setParseAction(lambda s, l, t: '~')
    not_regex = pp.Regex('!~|not regex').setParseAction(lambda s, l, t: '!~')
    any_op = pp.Regex('[^ ]*')
    operator = pp.Or([not_regex, regex, any_op]).setResultsName('operator')
    quote = pp.Suppress(pp.Optional("'"))
    value = (quote + pp.Regex('[^\']+', flags=re.I) + quote) \
        .setResultsName('value') \
        .setParseAction(lambda s, l, t: t[0])
    columns = pp.delimitedList(pp.Regex('[^, ]*'), delim=pp.Regex(', *')) \
        .setResultsName('display_columns')
    column = pp.Regex('[a-z]+', flags=re.I).setResultsName('column')
    conditional = column + operator + value
    head = select + columns + table
    grammar = head | conditional
    return grammar


def query_data(data, query):
    '''
    Parses SQL + regex query and applies it to given data.

    Regex operators:

        * ~, regex - Match regular expression
        * !~, not regex - Do not match regular expression

    Args:
        data (DataFrame): DataFrame to be queried.
        query (str): SQL query that may include regex operators.

    Returns:
        DataFrame: Data filtered by query.
    '''
    # split queries by where/and/or
    queries = re.split(' where | and | or ', query, flags=re.I)

    # detect whether any sub query has a regex operator
    has_regex = False
    for q in queries:
        if re.search(' regex | ~ | !~ ', q, flags=re.I):
            has_regex = True
            break

    # if no regex operator is found just submit query to pandasql
    if not has_regex:
        data = pandasql.sqldf(query, {'data': data})

    else:
        grammar = get_sql_grammar()

        # move select statement to end
        if 'select' in queries[0]:
            q = queries.pop(0)
            queries.append(q)

        for q in queries:
            # get column, operator and value
            parse = grammar.parseString(q).asDict()
            op = parse['operator']

            # initial select statement
            if op == 'select':
                data = pandasql.sqldf(q, {'data': data})

            # regex search
            elif op == '~':
                mask = data[parse['column']] \
                    .astype(str) \
                    .apply(lambda x: re.search(parse['value'], x, flags=re.I)) \
                    .astype(bool)
                data = data[mask]

            # regex not search
            elif op == '!~':
                mask = data[parse['column']] \
                    .astype(str) \
                    .apply(lambda x: re.search(parse['value'], x, flags=re.I)) \
                    .astype(bool)
                data = data[~mask]

            # ther SQL query
            else:
                data = pandasql.sqldf('select * from data where ' + q, {'data': data})

            if len(data) == 0:
                break
    return data


def query_dict(data, query):
    # type: (dict, str) -> dict
    '''
    Query a given diction with a given SQL query.

    Args:
        data (dict): Dictionary to be queried.
        query (str): SQL query.

    Returns:
        dict: Queried dictionary.
    '''
    data_ = data  # type: Any
    data_ = rpb.BlobETL(data_) \
        .to_flat_dict() \
        .items()
    data_ = DataFrame(list(data_), columns=['key', 'value'])
    data_ = query_data(data_, query)
    data_ = dict(zip(data_.key.tolist(), data_.value.tolist()))
    data_ = rpb.BlobETL(data_).to_dict()
    return data_



from setuptools import setup, find_packages

with open('version.txt') as f:
    VERSION = f.read().strip('\n')

with open('prod_requirements.txt') as f:
    PROD_REQUIREMENTS = f.read().split('\n')

with open('dev_requirements.txt') as f:
    DEV_REQUIREMENTS = f.read().split('\n')

with open('README.md') as f:
    README = f.read()
# ------------------------------------------------------------------------------

setup(
    name='shekels',
    packages=find_packages(where='./', exclude=['.*test.*.py', '.*.pyc']),
    package_dir={'shekels': 'shekels'},
    include_package_data=True,
    version=VERSION,
    license='MIT',
    description='Shekels is a local service which consumes a transactions CSV \
file downloaded from mint.intuit.com. It conforms this data into a database, \
and displays it as a searchable table and dashboard of configurable plots in \
web frontend.',
    long_description=README,
    long_description_content_type='text/markdown',
    author='Alex Braun',
    author_email='Alexander.G.Braun@gmail.com',
    url='https://github.com/theNewFlesh/shekels',
    download_url='https://github.com/theNewFlesh/shekels/archive/' + VERSION + '.tar.gz',
    keywords=[
        'dashboard', 'data', 'database', 'datastore', 'finance', 'flask',
        'json', 'plotly', 'service'
    ],
    install_requires=PROD_REQUIREMENTS,
    classifiers=[
      'Development Status :: 4 - Beta',
      'Intended Audience :: Developers',
      'License :: OSI Approved :: MIT License',
      'Programming Language :: Python :: 3',
      'Programming Language :: Python :: 3.7',
    ],
    extras_require={
        "dev": DEV_REQUIREMENTS
    },
)



from typing import Any, Dict, Union

import os
from pathlib import Path

import schematics.types as sty
from schematics.models import Model
from schematics.exceptions import CompoundError, DataError, ValidationError


# VALIDATORS--------------------------------------------------------------------
COLOR_SCHEME = dict(
    dark1='#040404',
    dark2='#141414',
    bg='#181818',
    grey1='#242424',
    grey2='#444444',
    light1='#A4A4A4',
    light2='#F4F4F4',
    dialog1='#444459',
    dialog2='#5D5D7A',
    red1='#F77E70',
    red2='#DE958E',
    orange1='#EB9E58',
    orange2='#EBB483',
    yellow1='#E8EA7E',
    yellow2='#E9EABE',
    green1='#8BD155',
    green2='#A0D17B',
    cyan1='#7EC4CF',
    cyan2='#B6ECF3',
    blue1='#5F95DE',
    blue2='#93B6E6',
    purple1='#C98FDE',
    purple2='#AC92DE',
)  # type: Dict[str, str]


def is_color_scheme(item):
    # type: (dict) -> None
    '''
    Determines if given dict is a valid color scheme.

    Args:
        item (dict): Color scheme dictionary.

    Raises:
        ValidationError: If item contains invalid keys.
    '''
    keys = list(COLOR_SCHEME.keys())
    ikeys = list(item.keys())

    diff = set(ikeys).difference(keys)  # type: Any
    diff = sorted(list(diff))
    if len(diff) > 0:
        msg = f'Invalid color scheme keys: {diff}.'
        raise ValidationError(msg)


def is_csv(filepath):
    # type: (Union[str, Path]) -> None
    '''
    Determines if given filepath is a CSV.

    Args:
        filepath (str or Path): Filepath.

    Raises:
        ValidationError: If filepath is not a CSV.
    '''
    filepath = Path(filepath).as_posix()
    ext = os.path.splitext(filepath)[-1][1:].lower()
    if not os.path.isfile(filepath) or ext != 'csv':
        msg = f'{filepath} is not a valid CSV file.'
        raise ValidationError(msg)


def is_comparator(item):
    # type: (str) -> None
    '''
    Ensures that given string is a legal comparator.

    Legal comparators:

        * ==
        * !=
        * >
        * >=
        * <
        * =<
        * ~
        * !~

    Args:
        item (str): String to be tested.

    Raises:
        ValidationError: If item is not a legal comparator.
    '''
    comps = ['==', '!=', '>', '>=', '<', '=<', '~', '!~']
    if item not in comps:
        msg = f'{item} is not a legal comparator. Legal comparators: {comps}.'
        raise ValidationError(msg)


def is_metric(item):
    # type: (str) -> None
    '''
    Ensures that given string is a legal metric.

    Legal metrics:

        * max
        * mean
        * min
        * std
        * sum
        * var
        * count

    Args:
        item (str): String to be tested.

    Raises:
        ValidationError: If item is not a legal metric.
    '''
    metrics = ['max', 'mean', 'min', 'std', 'sum', 'var', 'count']
    if item not in metrics:
        msg = f'{item} is not a legal metric. Legal metrics: {metrics}.'
        raise ValidationError(msg)


def is_plot_kind(item):
    '''
    Ensures item is a kind of plotly plot.

    Args:
        item (str): Kind of plot.

    Raises:
        ValidationError: If item is a plot kind.
    '''
    kinds = [
        'area', 'bar', 'barh', 'line' 'lines', 'ratio', 'scatter', 'spread'
    ]
    if item not in kinds:
        msg = f'{item} is not a legal plot kind. Legal kinds: {kinds}.'
        raise ValidationError(msg)


def is_bar_mode(item):
    '''
    Ensures mode is a legal bar mode.

    Args:
        item (str): Mode.

    Raises:
        ValidationError: If mode is not a legal bar mode.
    '''
    modes = ['stack', 'group', 'overlay']
    if item not in modes:
        msg = f'{item} is not a legal bar mode. Legal bar modes: {modes}.'
        raise ValidationError(msg)


def is_percentage(number):
    '''
    Ensures number is between 0 and 100.

    Args:
        number (float): Number to be tested.

    Raises:
        ValidationError: If number is not between 0 and 100.
    '''
    if number < 0 or number > 100:
        msg = f'{number} is not a legal percentage. '
        msg += f'{number} is not between 0 and 100.'
        raise ValidationError(msg)


# SCHEMATICS--------------------------------------------------------------------
class FilterAction(Model):
    '''
    Schematic for filter actions.

    Attributes:
        column (str): Column name.
        comparator (str): String representation of comparator.
        value (object): Value to be compared.
    '''
    column = sty.StringType(required=True)
    comparator = sty.StringType(required=True, validators=[is_comparator])
    value = sty.BaseType(required=True)


class GroupAction(Model):
    '''
    Schematic for group actions.

    Attributes:
        columns (str or list[str]): Columns to group data by.
        metric (str): Aggregation metric.
        datetime_column (str, optinal): Datetime column for time grouping.
            Default: date.
    '''
    columns = sty.ListType(sty.StringType(), required=True)
    metric = sty.StringType(required=True, validators=[is_metric])
    datetime_column = sty.StringType(required=True, default='date')


class PivotAction(Model):
    '''
    Schematic for group actions.

    Attributes:
        columns (list[str]): Columns whose unique values become separate traces
            within a plot.
        values (list[str], optional): Columns whose values become the values
            within each trace of a plot. Default: [].
        index (str, optional): Column whose values become the y axis values of a
            plot. Default: None.
    '''
    columns = sty.ListType(sty.StringType(), required=True)
    values = sty.ListType(sty.StringType(), required=True, default=[])
    index = sty.StringType(required=True, default=None)


class ConformAction(Model):
    '''
    Schematic for conform actions.

    Attributes:
        action (str): Must be 'overwrite' or 'substitute'.
        source_column (str): Source column to be matched.
        target_column (str): Target column to be overwritten.
        mapping (dict): Mapping of matched key in source column with replacement
            value in target column.
    '''
    action = sty.StringType(required=True, choices=['overwrite', 'substitute'])
    source_column = sty.StringType(required=True)
    target_column = sty.StringType(required=True)
    mapping = sty.DictType(
        sty.UnionType(
            types=[sty.FloatType, sty.IntType, sty.BooleanType, sty.StringType]
        ),
        required=True,
    )

    def validate(self):
        '''
        Validates the state of the model. If the data is invalid, raises a
        DataError with error messages. Also, performs a stricter validation on
        mapping if action is substitute.

        Args:
            partial (bool, optional): Allow partial data to validate.
                Essentially drops the required=True settings from field
                definitions. Default: False.
            convert (bool, optional): Controls whether to perform import
                conversion before validating. Can be turned off to skip an
                unnecessary conversion step if all values are known to have the
                right datatypes (e.g., when validating immediately after the
                initial import). Default: True.

        Raises:
            DataError: If data is invalid.
        '''
        super().validate()
        if self.action == 'substitute':
            try:
                sty.DictType(sty.StringType(), required=True)\
                    .validate(self.mapping)
            except CompoundError as e:
                raise DataError(e.to_primitive())


class FigureItem(Model):
    '''
    Schematic for a plot figure.

    Attributes:
        kind (str): Type of plot. Default: bar.
        color_scheme (dict[str, str]): Color scheme for plot.
            Default: {'grey1': '#181818', 'bg': '#242424'}.
        x_axis (str): Column to use as x axis: Default: None.
        y_axis (str): Column to use as y axis: Default: None.
        title (str): Title of plot. Default: None.
        x_title (str): Title of plot x axis. Default: None.
        y_title (str): Title of plot y axis. Default: None.
        bins (int): Number of bins if histogram. Default: 50.
        bar_mode (str): How bars in bar graph are presented. Default: stack.
    '''
    kind = sty.StringType(default='bar', validators=[is_plot_kind])
    color_scheme = sty.DictType(
        sty.StringType(), default=dict(grey1='#181818', bg='#242424')
    )
    x_axis = sty.StringType(default=None)
    y_axis = sty.StringType(default=None)
    title = sty.StringType(default=None)
    x_title = sty.StringType(default=None)
    y_title = sty.StringType(default=None)
    bins = sty.IntType(default=50)
    bar_mode = sty.StringType(validators=[is_bar_mode], default='stack')


class PlotItem(Model):
    '''
    Schematic for a plot.

    Attributes:
        filters (list[dict]): How data is filtered. Default: [].
        group (dict): How data is grouped. Default: {}.
        pivot (dict): How data is pivoted. Default: {}.
        figure (dict): Plot figure details. Default: {}.
        min_width (float): Minimum width of plot. Default: 0.25.
    '''
    filters = sty.ListType(sty.ModelType(FilterAction), default=[])
    group = sty.ModelType(GroupAction)
    pivot = sty.ModelType(PivotAction)
    figure = sty.ModelType(FigureItem, default={})
    min_width = sty.FloatType(default=25, validators=[is_percentage])


class Config(Model):
    '''
    Configuration of database.

    Attributes:
        data_path (str): Path to CSV file.
        columns (list[str]): Columns to be displayed in data.
        default_query (str): Placeholder SQL query string.
        font_family (str): Font family.
        color_scheme (dict): Color scheme.
        conform (lit[dict]): List of conform actions.
        plots (list[dict]): List of plots.
    '''
    data_path = sty.StringType(required=True, validators=[is_csv])
    columns = sty.ListType(sty.StringType, default=[])
    default_query = sty.StringType(default='select * from data')
    font_family = sty.StringType(default='sans-serif, "sans serif"')
    color_scheme = sty.DictType(
        sty.StringType(), validators=[is_color_scheme], default=COLOR_SCHEME
    )
    conform = sty.ListType(sty.ModelType(ConformAction), default=[])
    plots = sty.ListType(sty.ModelType(PlotItem), default=[])



from typing import Any, Dict, Optional

from copy import deepcopy
from pprint import pformat
import base64
import json
import os
import re
import traceback

from dash.exceptions import PreventUpdate
from schematics.exceptions import DataError
import dash
import flask
import jinja2
import jsoncomment as jsonc
import lunchbox.tools as lbt
import rolling_pin.blob_etl as rpb

import shekels.core.config as cfg
import shekels.core.data_tools as sdt
import shekels.server.components as svc
# ------------------------------------------------------------------------------


TEMPLATE_DIR = lbt.relative_path(__file__, '../templates').as_posix()
if 'REPO_ENV' in os.environ.keys():
    TEMPLATE_DIR = lbt.relative_path(__file__, '../../../templates').as_posix()


def error_to_response(error):
    # type: (Exception) -> flask.Response
    '''
    Convenience function for formatting a given exception as a Flask Response.

    Args:
        error (Exception): Error to be formatted.

    Returns:
        flask.Response: Flask response.
    '''
    args = []  # type: Any
    for arg in error.args:
        if hasattr(arg, 'items'):
            for key, val in arg.items():
                args.append(pformat({key: pformat(val)}))
        else:
            args.append(str(arg))
    args = ['    ' + x for x in args]
    args = '\n'.join(args)
    klass = error.__class__.__name__
    msg = f'{klass}(\n{args}\n)'
    return flask.Response(
        response=json.dumps(dict(
            error=error.__class__.__name__,
            args=list(map(str, error.args)),
            message=msg,
            code=500,
            traceback=traceback.format_exc(),
        )),
        mimetype='application/json',
        status=500,
    )


def render_template(filename, parameters, directory=None):
    # type: (str, Dict[str, Any], Optional[str]) -> bytes
    '''
    Renders a jinja2 template given by filename with given parameters.

    Args:
        filename (str): Filename of template.
        parameters (dict): Dictionary of template parameters.
        directory (str or Path, optional): Templates directory.
            Default: '../../../templates'.

    Returns:
        bytes: HTML.
    '''
    tempdir = TEMPLATE_DIR
    if directory is not None:
        tempdir = lbt.relative_path(__file__, directory).as_posix()

    env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(tempdir),
        keep_trailing_newline=True
    )
    output = env.get_template(filename).render(parameters).encode('utf-8')
    return output


def parse_json_file_content(raw_content):
    # type: (str) -> Dict
    '''
    Parses JSON file content as supplied by HTML request.

    Args:
        raw_content (str): Raw JSON file content.

    Raises:
        ValueError: If header is invalid.
        JSONDecodeError: If JSON is invalid.

    Returns:
        dict: JSON content or reponse dict with error.
    '''
    header, content = raw_content.split(',')
    temp = header.split('/')[-1].split(';')[0]  # type: str
    if temp != 'json':
        msg = f'File header is not JSON. Header: {header}.'
        raise ValueError(msg)

    output = base64.b64decode(content).decode('utf-8')
    return jsonc.JsonComment().loads(output)


def update_store(client, store, endpoint, data=None):
    # type (FlaskClient, dict, str, Optional(dict)) -> None
    '''
    Updates store with data from given endpoint.
    Makes a post call to endpoint with client.

    Args:
        client (FlaskClient): Flask client instance.
        store (dict): Dash store.
        endpoint (str): API endpoint.
        data (dict, optional): Data to be provided to endpoint request.
    '''
    if data is not None:
        store[endpoint] = client.post(endpoint, json=json.dumps(data)).json
    else:
        store[endpoint] = client.post(endpoint).json


def store_key_is_valid(store, key):
    # type: (dict, str) -> bool
    '''
    Determines if given key is in store and does not have an error.

    Args:
        store (dict): Dash store.
        key (str): Store key.

    Raises:
        PreventUpdate: If key is not in store.

    Returns:
        bool: True if key exists and does not have an error key.
    '''
    if key not in store:
        raise PreventUpdate
    value = store[key]
    if isinstance(value, dict) and 'error' in value:
        return False
    return True


def solve_component_state(store, config=False):
    # type (dict) -> Optional(html.Div)
    '''
    Solves what component to return given the state of the given store.

    Returns a key value table component embedded with a relevant message or error
    if a required key is not found in the store, or it contain a dictionary with
    am "error" key in it. Those required keys are as follows:

        * /config
        * /config/search
        * /api/initialize
        * /api/update
        * /api/search

    Args:
        store (dict): Dash store.
        config (bool, optional): Whether the component is for the config tab.
            Default: False.

    Returns:
        Div: Key value table if store values are not present or have errors,
            otherwise, none.
    '''
    states = [
        ['/config', None],
        ['/config/search', None],
        ['/api/initialize', 'Please call init or update.'],
        ['/api/update', 'Please call update.'],
        ['/api/search', None],
    ]
    if config:
        states = states[:2]
        states[1][1] = None
    for key, message in states:
        value = store.get(key)
        if message is not None and value is None:
            return svc.get_key_value_table(
                {'action': message},
                id_='status',
                header='status',
            )
        elif isinstance(value, dict) and 'error' in value:
            return svc.get_key_value_table(
                value,
                id_='error',
                header='error',
                key_order=['error', 'message', 'code', 'traceback'],
            )
    return None


# EVENTS------------------------------------------------------------------------
def config_query_event(value, store, app):
    # type: (str, dict, dash.Dash) -> dict
    '''
    Updates given store given a config query.

    Args:
        value (str): SQL query.
        store (dict): Dash store.
        app (dash.Dash): Dash app.

    Returns:
        dict: Modified store.
    '''
    value = value or 'select * from config'
    value = re.sub('from config', 'from data', value, flags=re.I)
    try:
        store['/config/search'] = sdt.query_dict(app.api.config, value)
    except Exception as e:
        store['/config/search'] = error_to_response(e).json
    return store


def config_edit_event(value, store, app):
    # type: (dict, dict, dash.Dash) -> dict
    '''
    Saves given edits to store.

    Args:
        value (dict): Config table.
        store (dict): Dash store.
        app (dash.Dash): Dash app.

    Returns:
        dict: Modified store.
    '''
    new = value['new']
    old_key = value['old']['key']
    config = store.get('/config', deepcopy(app.api.config))
    items = [
        ('/config', config),
        ('/config/search', store.get('/config/search', config)),
    ]
    for key, val in items:
        item = rpb.BlobETL(val).to_flat_dict()
        if old_key in item:
            del item[old_key]
            item[new['key']] = new['value']
        store[key] = rpb.BlobETL(item).to_dict()
    return store


def data_query_event(value, store, app):
    # type: (str, dict, dash.Dash) -> dict
    '''
    Updates given store given a data query.

    Args:
        value (str): SQL query.
        store (dict): Dash store.
        app (dash.Dash): Dash app.

    Returns:
        dict: Modified store.
    '''
    update_store(app.client, store, '/api/search', data={'query': value})
    store['/api/search/query'] = value
    return store


def init_event(value, store, app):
    # type: (None, dict, dash.Dash) -> dict
    '''
    Initializes app database.

    Args:
        value (None): Ignored.
        store (dict): Dash store.
        app (dash.Dash): Dash app.

    Returns:
        dict: Modified store.
    '''
    update_store(app.client, store, '/api/initialize', data=app.api.config)
    if 'error' in store['/api/initialize']:
        store['/config'] = store['/api/initialize']
    else:
        store['/config'] = deepcopy(app.api.config)
    return store


def update_event(value, store, app):
    # type: (None, dict, dash.Dash) -> dict
    '''
    Update app database.

    Args:
        value (None): Ignored.
        store (dict): Dash store.
        app (dash.Dash): Dash app.

    Returns:
        dict: Modified store.
    '''
    update_store(app.client, store, '/api/update')
    update_store(
        app.client,
        store,
        '/api/search',
        data={'query': app.api.config['default_query']}
    )
    return store


def upload_event(value, store, app):
    # type: (str, dict, dash.Dash) -> dict
    '''
    Uploads config to app store.

    Args:
        value (str): Config.
        store (dict): Dash store.
        app (dash.Dash): Dash app.

    Returns:
        dict: Modified store.
    '''
    try:
        config = parse_json_file_content(value)
        config = cfg.Config(config)
        config.validate()
        store['/config'] = config.to_primitive()
        store['/config/search'] = deepcopy(store['/config'])
    except Exception as error:
        store['/config/search'] = error_to_response(error).json
    return store


def save_event(value, store, app):
    # type: (None, dict, dash.Dash) -> dict
    '''
    Save store config to app.api.config path.

    Args:
        value (None): Ignore me.
        store (dict): Dash store.
        app (dash.Dash): Dash app.

    Returns:
        dict: Modified store.
    '''
    try:
        config = store.get('/config', app.api.config)
        cfg.Config(config).validate()
        with open(app.api.config_path, 'w') as f:
            json.dump(config, f, indent=4, sort_keys=True)
    except (Exception, DataError) as error:
        store['/config'] = deepcopy(app.api.config)
        store['/config/search'] = error_to_response(error).json
    return store



from typing import Any, List

from lunchbox.enforce import Enforce, EnforceError
from pandas import DataFrame
# ------------------------------------------------------------------------------


def enforce_dataframes_are_equal(a, b):
    '''
    Endsures that DataFrames a and b have equal contents.

    Args:
        a (DataFrame): DataFrame A.
        b (DataFrame): DataFrame B.

    Raises:
        EnforceError: If a and b are not equal.
    '''
    # column names
    a_cols = set(a.columns.tolist())
    b_cols = b.columns.tolist()
    diff = a_cols.symmetric_difference(b_cols)
    diff = sorted(list(diff))

    msg = f'A and b have different columns: {diff}.'
    Enforce(len(diff), '==', 0, message=msg)

    # shape
    msg = 'A and b have different shapes. {a} != {b}.'
    Enforce(a.shape, '==', b.shape, message=msg)

    # NaNs cannot be compared
    a = a.fillna('---NAN---')
    b = b.fillna('---NAN---')

    # values
    errors = []
    for col in a.columns:
        mask = a[col] != b[col]
        a_vals = a.loc[mask, col].tolist()
        if len(a_vals) > 0:
            b_vals = b.loc[mask, col].tolist()
            error = [[col, av, bv] for av, bv in zip(a_vals, b_vals)]
            errors.extend(error)

    if len(errors) > 0:
        msg = DataFrame(errors, columns=['column', 'a', 'b']).to_string()
        msg = f'DatFrames have different values:\n{msg}'
        raise EnforceError(msg)

    #  records
    a = a.to_dict(orient='records')
    b = b.to_dict(orient='records')
    Enforce(a, '==', b)


def enforce_columns_in_dataframe(columns, data):
    # type: (List[str], DataFrame) -> None
    '''
    Ensure all given columns are in given dataframe columns.

    Args:
        columns (list[str]): Column names.
        data (DataFrame): DataFrame.

    Raises:
        EnforceError: If any column not found in data.columns.
    '''
    cols = data.columns.tolist()
    diff = set(columns).difference(cols)  # type: Any
    diff = sorted(list(diff))
    msg = f'Given columns not found in data. {diff} not in {cols}.'
    Enforce(diff, '==', [], message=msg)
